#include "stdafx.h"
#include "tse.h"

bool isdiag=false;
bool issilent=false;

void showerrmsg(t_tserror& tserror)
{
	if (issilent) return;
	cout<<"\nError "<<tserror.tserrcode<<": ";
	cout<<tserrdescr[tserror.tserrcode]<<" ";
	cout<<tserror.errinfo.c_str();
	cout<<"\n";
}

void showdiagmsg(const char* sdiag, bool donewline)
{
	if ((!isdiag) || (issilent)) return;
	cout<<sdiag;
	if (donewline) cout<<"\n";
}


c_telnet::c_telnet()
{
	WSADATA lpWSAData;

	session_open=false;
	showdiagmsg(tsdiagdescr[tsdiag_initwinsock],true);
	WSAStartup(0x202,&lpWSAData);
	ipaddr="";
	tcpport=telnet_port;
	sctadr.sin_family=AF_INET;
}

c_telnet::~c_telnet()
{
	close_session();
	showdiagmsg(tsdiagdescr[tsdiag_closewinsock],true);
	WSACleanup();
}


bool c_ts::is_arg_param(char* arg)
{
	if (strchr(tsoptin_prefix,arg[0])!=NULL)
	{
		if (arg[0]==arg[1]) return true;
		else return false;
	}
	else return true;
}


c_ts::c_ts(int argc, char* argv[])
{
	tsaction=tsrun;
	tsoption.tsdiag=0;
	tsoption.tssilent=0;
	oneline="";
	onelinetmp="";
	linenumber=0;

	showdiagmsg(tsdiagdescr[tsdiag_transparam],true);
	for(int i=1;i<argc;i++)
	{
		if (!is_arg_param(argv[i])) 
		{	
			if (strlen(argv[i])!=2) 
			{
				tsaction=tserror;
				throw t_tserror(tsinvalid_option,argv[i]);
			}
			switch ((t_tsoption_letter) argv[i][1])
			{
			case tslet_help1:
			case tslet_help2:
				tsaction=tshelp;
				break;
			case tslet_diag:
				tsoption.tsdiag=1;
				isdiag=true;
				break;
			case tslet_silent:
				tsoption.tssilent=1;
				issilent=true;
				break;
			default:
				tsaction=tserror;
				throw t_tserror(tsinvalid_option,argv[i]);
				break;
			}
		}
	}
	if (tsaction==tshelp) return;

	tsparam.resize(9);
	int param_count=0;
	char* parg=NULL;
	for(i=1;i<argc;i++)
	{
		if (is_arg_param(argv[i])) 
		{
			parg=argv[i];
			if (strchr(tsoptin_prefix,argv[i][0])!=NULL) parg++;
			if (param_count<9)
			{
				tsparam[param_count]=parg;
				param_count++;
			}
		}
	}
}


void c_ts::error_in_script(t_tserrcode errcode)
{
	char errinfo[maxinputstring];
	sprintf(errinfo,"\nLine %i. ",linenumber);
	strcat(errinfo,oneline.c_str());
	throw t_tserror(errcode,errinfo);
}

void c_ts::getoneline()
{
	char buffer[maxinputstring];
	oneline="";
	showdiagmsg(tsdiagdescr[tsdiag_getnextline],true);
	cin.getline(buffer,maxinputstring);
	oneline=buffer;
	linenumber++;
	showdiagmsg(buffer,true);
}


void ltrim (string& s)
{
	int pos=s.find_first_not_of (tsemptysymbols);
	if (pos==s.npos) s="";
	else s.erase(0,pos) ;
}

void rtrim (string& s)
{
	int pos=s.find_last_not_of (tsemptysymbols);
	if (pos==s.npos) s="";
	else s.erase(pos+1,s.length()-pos-1) ;
}

void alltrim (string& s)
{
	ltrim(s);
	rtrim(s);
}


void subst_str(string& s, const char* ps_change, const char* ps_new)
{
	int oldpos=0;
	int newpos=0;

	while ((newpos+1)<s.length())
	{
		oldpos=newpos;
		newpos=s.find(ps_change,oldpos);
		if ((newpos!=oldpos) && (newpos!=s.npos))
		{
			if (newpos!=0)
				if (s[newpos-1]==ps_change[0]) 
				{
					s.erase(newpos,1) ;
					continue;
				}
			s.replace(newpos,strlen(ps_change),ps_new);
			newpos=newpos+strlen(ps_new);
		}
		else break;
	}
}


void c_ts::processline()
{
	int i=0;
	char param_str[3];

	showdiagmsg(tsdiagdescr[tsdiag_processline],true);
	alltrim(oneline);	
	if (oneline.length()==0) return;
	if (oneline[0]==tsruncommand_comment)
	{
		oneline="";
		return;
	}

	param_str[0]=tssubst_param;
	param_str[2]=0;
	for (i=0;i<9;i++)
	{
		param_str[1]='1'+i;
		subst_str(oneline,param_str,tsparam[i].c_str());
	}

	for (i=0;i<tssubst_table_size;i++)
		subst_str(oneline,tssubst_table[i].sinit, tssubst_table[i].send);

	alltrim(oneline);
}


void delay (int seconds)
{
	clock_t timestart;
	clock_t timecurr;
	double duration;

	timestart=clock();
	while (true)
	{
		timecurr=clock();
		duration = (double)(timecurr- timestart) / CLOCKS_PER_SEC;
   		if (duration >seconds) break;
	}
}



void c_telnet::close_session()
{
	if (session_open) {
		showdiagmsg(tsdiagdescr[tsdiag_closesession],true);	
		linger lng;
		lng.l_onoff=-1; lng.l_linger=0;
		delay(1);
		setsockopt(sct,SOL_SOCKET,SO_LINGER, (const char*) &lng, sizeof(lng));
		if (closesocket(sct)) return;
	}
	session_open=false;
}


void c_telnet::closewait_session()
{
	if (session_open) {
		showdiagmsg(tsdiagdescr[tsdiag_closesession],true);
		linger lng;
		lng.l_onoff=-1; lng.l_linger=maxtime_waitclose;
		delay(1);
		setsockopt(sct,SOL_SOCKET,SO_LINGER, (const char*) &lng, sizeof(lng));
		if (closesocket(sct)) return;
	}
	session_open=false;
}


bool c_telnet::open_session(string ipadr, unsigned int port)
{
	HOSTENT* host;
	unsigned long ipnumaddr;
	
	ipaddr=ipadr;
	tcpport=port;
	close_session();
	sctadr.sin_port=htons(tcpport);
	if (isalpha(ipaddr[0])) 
	{
		showdiagmsg(tsdiagdescr[tsdiag_gethost],true);
		host=gethostbyname(ipaddr.c_str());
		if (host==NULL)	return false;	
		memcpy(&(sctadr.sin_addr),host->h_addr,sizeof(ipnumaddr));
	}
	else 
	{
		ipnumaddr = inet_addr(ipaddr.c_str());
		memcpy(&(sctadr.sin_addr),&ipnumaddr,sizeof(ipnumaddr));
	}
	showdiagmsg(tsdiagdescr[tsdiag_opensocket],true);	
	sct=socket(AF_INET,SOCK_STREAM,0);
	if (sct==INVALID_SOCKET) return false;
	session_open=true;
	u_long ul=0;
	ioctlsocket(sct,FIONBIO,&ul);

	showdiagmsg(tsdiagdescr[tsdiag_connecthost],false);	
	showdiagmsg(ipaddr.c_str(),true);	
	if (connect(sct,(struct sockaddr*) &sctadr,sizeof(sctadr))==SOCKET_ERROR) 
	{
		close_session();
		return false;
	}

	return session_open;
}


bool c_telnet::getwait_nextchar()
{
	if (recv(sct,&chinout,1, 0)==1) {
		if (!issilent) cout<<chinout;
		return true;
	}
	else return true;
}

bool c_telnet::put_nextchar(char ch)
{
	chinout=ch;
	if (send(sct,&chinout,1,0)==1) {
		if (!issilent) cout<<chinout;
		return true;
	}
	else return false;
}
   


void extract_word(const string s, string& sword, string& srest)
{
	int pos=s.find_first_of(tsemptysymbols);
	if (pos==s.npos)
	{
		sword=s;
		srest="";
	}
	else 
	{
		sword=s.substr(0,pos);
		srest=s;
		srest.erase(0,pos);
	}

}

void c_ts::extract_runcommand()
{
	char command[maxinputstring];
	string sword;

	extract_word(oneline,sword,onelinetmp);
	ltrim(onelinetmp);
	strcpy(command,sword.c_str());
	_strlwr(command);

	for (tsruncommand=tsr_pause;tsruncommand!=tsr_error;tsruncommand=static_cast <t_tsruncommand> (tsruncommand-1))
		if (strcmp(tsruncommands[tsruncommand], command)==0) break;
}


void c_ts::dopause()
{
	int pause_time=0;
	int res=sscanf(onelinetmp.c_str(),"%i",&pause_time);
	if ((pause_time<=0) || (res==0) || (res==EOF)) error_in_script(tsbad_pause_time);
	showdiagmsg(tsdiagdescr[tsdiag_pause],true);	
	delay(pause_time);
}


void c_ts::preparetextline(t_tserrcode errcode)
{
	if (onelinetmp.length()<2) error_in_script(errcode);
	if ((onelinetmp[0]!='"') || onelinetmp[onelinetmp.length()-1]!='"') error_in_script(errcode);
	onelinetmp.erase(onelinetmp.length()-1,1);
	onelinetmp.erase(0,1);

}

void c_ts::sendtext()
{
	showdiagmsg(tsdiagdescr[tsdiag_sendtext],false);	
	preparetextline(tsbad_sendline_format);
	showdiagmsg(onelinetmp.c_str(),true);	
	for (int i=0;i<onelinetmp.length();i++)
		if (!put_nextchar(onelinetmp[i])) throw t_tserror(tserr_send_char,onelinetmp.c_str());
}


void c_ts::waittext()
{
	showdiagmsg(tsdiagdescr[tsdiag_waittext],false);	
	showdiagmsg(onelinetmp.c_str(),true);	
	preparetextline(tsbad_waitline_format);
	if (onelinetmp.length()==0) return;
	string waitbuf="";
	while (true)
	{
		if (!getwait_nextchar()) throw t_tserror(tserr_get_char,"");
		if (waitbuf.length()==onelinetmp.length()) waitbuf.erase(0,1);
		waitbuf.append(1, chinout);
		if (strcmp(waitbuf.c_str(),onelinetmp.c_str())==0) break;
	}
}


void c_ts::connecttohost()
{
	string portstr;
	string ipadr;
	unsigned int port=0;

	extract_word(onelinetmp, ipadr, portstr);
	if (portstr.length()==0) port=telnet_port
	else	sscanf(portstr.c_str(),"%u",&port);
	if (!open_session(ipadr, port)) throw t_tserror(tserr_open_session,onelinetmp.c_str());
}


void c_ts::executeline() 
{
	extract_runcommand();
	showdiagmsg(tsdiagdescr[tsdiag_execline],true);	
	switch (tsruncommand)
	{
	case tsr_error:
		error_in_script(tsunknow_command);
		break;
	case tsr_none:
		break;
	case tsr_open:
		connecttohost();
		break;
	case tsr_close:
		closewait_session();
		break;
	case tsr_exit:
		closewait_session();
		quitflag=true;
		break;
	case tsr_quit:
		close_session();
		quitflag=true;
		break;
	case tsr_wait:
		waittext();
		break;
	case tsr_send:
		sendtext();
		break;
	case tsr_pause:
		dopause();
		break;
	}
}


void c_ts::run()
{
	if (!issilent) cout<<ts_copyright;
	if (tsaction==tshelp)  { if (!issilent) 	cout<<ts_helpmsg;}
	else
	{
	quitflag=false;
	while (!cin.eof() && !quitflag)
	{
		getoneline();
		processline();
		if (oneline.length()!=0) executeline();
	}
	}
}


int main(int argc, char* argv[])
{
	try
	{
		c_ts ts(argc,argv);
		ts.run();
		return tssuccess;
	}
	catch (t_tserror tserror)
	{
		showerrmsg(tserror);
		return tserror.tserrcode;
	}
}
